"""
handles access to common resources
"""
from functools import partial
import pkg_resources

ASCII_ART = 'resources/ascii_art.txt'
DEFAULT_CONFIG = 'resources/config_default.yml'
DEFAULT_LABELS = 'resources/labels_default.yml'
ISSUE_TEMPLATE = 'resources/issue_template.yml'

get_default_config = partial(pkg_resources.resource_string,
                             __name__, DEFAULT_CONFIG)
get_default_labels = partial(pkg_resources.resource_string,
                             __name__, DEFAULT_LABELS)
get_ascii_art = partial(pkg_resources.resource_string,
                        __name__, ASCII_ART)
get_issue_template = partial(pkg_resources.resource_string,
                             __name__, ISSUE_TEMPLATE)



from __future__ import print_function

import getpass
import warnings

import attr
import yaml
from jira import JIRA
from jira.exceptions import JIRAError

from .common import (
    iso_time_is_today, sanitize_worklog_time, friendly_worklog_time, iso_time_is_yesterday)


class InvalidLabelError(Exception):
    def __init__(self, component, label, *args, **kwargs):
        super(InvalidLabelError, self).__init__(*args, **kwargs)
        self.component = component
        self.label = label

    def __str__(self):
        return "Label '{}' is not valid for component '{}'".format(self.label, self.component)


class JiraClientOverride(JIRA):
    def _create_kerberos_session(self, *args, **kwargs):
        """
        Little hack to get auth cookies from JIRA when using kerberos, otherwise
        queries to other URLs hit a 401 and are not handled properly for some
        reason

        https://stackoverflow.com/questions/21578699/jira-rest-api-and-kerberos-authentication
        """
        super(JiraClientOverride, self)._create_kerberos_session(*args, **kwargs)
        print("Attempting to authenticate with kerberos...")
        r = self._session.get("{}/step-auth-gss".format(self._options['server']))
        if r.status_code == 200:
            print("Authenticated successfully")


class IssueFields(object):
    """
    Class which holds builders for various jira field data

    Build multiple field sections and then pass in the entire kwarg without having to remember the
    json variations.
    Example:

    f = IssueFields().labels(['something1', 'something2']).component("LOL").summary("my summary")
    issue.update(**f.kwarg)
    """
    def __init__(self):
        self._base = {'fields': {}}
        self.fields = self._base['fields']

    @property
    def kwarg(self):
        return self._base

    def timetracking(self, remaining, original):
        self.fields.update({
            'timetracking': {
                'remainingEstimate': sanitize_worklog_time(str(remaining)),
                'originalEstimate': sanitize_worklog_time(str(original))
            }
        })
        return self

    def component(self, component_name):
        if component_name:
            self.fields.update({
                'components': [
                    {'name': component_name}
                ]
            })
        return self

    def labels(self, label_list):
        if label_list:
            self.fields.update({
                'labels': label_list
            })
        return self

    def summary(self, summary_text):
        if summary_text:
            self.fields.update({
                'summary': summary_text
            })
        return self

    def description(self, description_text):
        if description_text:
            self.fields.update({
                'description': description_text
            })
        return self

    def assignee(self, name):
        if name:
            self.fields.update({
                'assignee': {
                    'name': name
                }
            })
        return self

    def issuetype(self, name):
        if name:
            self.fields.update({
                'issuetype': {
                    'name': name
                }
            })
        return self

    def project(self, name=None, key=None, id=None):
        kwargs = {'name': name, 'key': key, 'id': id}
        if not any(kwargs.values()):
            raise ValueError("project needs at least 1 of: [name, key, id] defined")

        d = {'project': {}}    

        for key, value in kwargs.items():
            if value:
                d['project'][key] = value

        self.fields.update(d)
        return self

@attr.s
class JiraWrapper(object):
    """
    Provides utils for storing config and interacting with python-jira
    """
    config_file = attr.ib()
    labels_file = attr.ib()

    _config = attr.ib(default=attr.Factory(dict))
    _component_labels_map = attr.ib(default=attr.Factory(dict))
    _jira = attr.ib(default=None)
    _current_sprint_id = attr.ib(default=0)
    _current_sprint_name = attr.ib(type=str, default=None)
    _board_id = attr.ib(default=0)
    _project_id = attr.ib(default=0)
    _userid = attr.ib(type=str, default=None)

    def __attrs_post_init__(self):
        """
        After object instantiation, load config files
        """
        with open(self.config_file, 'r') as f:
            self._config = yaml.safe_load(f)
        if self.labels_file:
            with open(self.labels_file, 'r') as f:
                self._component_labels_map = yaml.safe_load(f)

    @property
    def jira_url(self):
        """
        Server URL being used for JIRA.
        """
        return self._config['url']

    @property
    def label_check(self):
        try:
            return self._config['label_check']
        except KeyError:
            return False

    @property
    def verify_ssl(self):
        try:
            return self._config['verify_ssl']
        except KeyError:
            return True

    @property
    def jira(self):
        """
        Creates the JiraClient session
        """
        if not self._jira:
            print("Connecting to jira at", self.jira_url)
            kwargs = {}
            cfg = self._config
            auth_cfg = cfg['auth']
            kwargs['validate'] = False

            if 'basic_auth' in auth_cfg and auth_cfg['basic_auth'] is True:
                print("Using basic authentication")
                if 'password' in auth_cfg and auth_cfg['password']:
                    password = auth_cfg['password']
                else:
                    password = getpass.getpass("Enter your JIRA password: ")
                kwargs['basic_auth'] = (auth_cfg['username'], password)
            else:
                print("Using kerberos authentication")
                kwargs['kerberos'] = True
                kwargs['kerberos_options'] = {'mutual_authentication': "DISABLED"}

            kwargs['options'] = {'server': self.jira_url}
            if 'ca_cert_path' in self._config:
                kwargs['options']['verify'] = self._config['ca_cert_path']
            if self.verify_ssl is False:
                print("Warning: SSL certificate verification is disabled!")
                kwargs['options']['verify'] = False
                # Disable ssl validation warnings, we gave one warning already ...
                from urllib3.exceptions import InsecureRequestWarning
                from requests.packages.urllib3 import disable_warnings
                disable_warnings(category=InsecureRequestWarning)

            self._jira = JiraClientOverride(**kwargs)
        return self._jira

    @property
    def board_id(self):
        if not self._board_id:
            try:
                cfgboard = str(self._config['board']).lower()
            except KeyError:
                raise KeyError("config has no 'board' defined!")

            boards = self.jira.boards()

            for b in boards:
                if b.name.lower() == cfgboard or str(b.id) == cfgboard:
                    self._board_id = str(b.id)
                    break

            if not self._board_id:
                raise ValueError("Unable to find board '{}'".format(self._config['board']))
        return self._board_id

    @property
    def project_id(self):
        if not self._project_id:
            try:
                cfgproject = str(self._config['project']).lower()
            except KeyError:
                raise KeyError("config has no 'project' defined!")

            projects = self.jira.projects()

            for p in projects:
                if any(x == cfgproject for x in [p.key.lower(), p.name.lower(), str(p.id)]):
                    self._project_id = str(p.id)
                    break

            if not self._project_id:
                raise ValueError("Unable to find project '{}'".format(self._config['project']))
        return self._project_id

    @property
    def userid(self):
        if not self._userid:
            self._userid = self.jira.myself()['key']

        return self._userid

    def find_sprint(self, txt):
        """
        Return sprint ID whose name or ID contains "txt", case insensitive.

        Args:
          txt: string or int

        Returns:
          tuple of (sprint_name, sprint_id)
        """
        sprints = self.jira.sprints(board_id=self.board_id)
        for s in sprints:
            txt = str(txt).lower()
            if txt.isdigit() and txt in [n for n in s.name.split() if n.isdigit()]:
                return s.name, str(s.id)
            elif not txt.isdigit() and txt in s.name.lower():
                return s.name, str(s.id)
        raise ValueError("Unable to find sprint with text: ", str(txt))

    def get_current_sprint(self):
        active_sprints = (
            sprint for sprint
            in self.jira.sprints(board_id=self.board_id, state='active')
            if sprint.state.lower() == 'active'
        )
        current_sprint = sorted(active_sprints, key=lambda sprint: sprint.id)[-1]
        self._current_sprint_id = str(current_sprint.id)
        self._current_sprint_name = current_sprint.name
        return current_sprint

    @property
    def current_sprint_id(self):
        """
        Returns currently active sprint ID for the agile board.
        """
        if not self._current_sprint_id:
            self.get_current_sprint()
        return self._current_sprint_id

    @property
    def current_sprint_name(self):
        if not self._current_sprint_name:
            self.get_current_sprint()
        return self._current_sprint_name

    def search_issues(self, assignee=None, sprint=None, status=None, text=None):
        """
        Search issues

        Args:
           sprint: sprint ID number, sprint name, or "backlog", default is current sprint
           assignee: user id, default is "currentUser"
           status: for e.x. "in progress"

        Returns:
            List of JIRA.Issue resources
        """
        if sprint == "backlog":
            search_query = (
                "project = {} AND issuetype != Epic AND resolution = Unresolved AND "
                "status != Done AND "
                "(Sprint = EMPTY OR Sprint not in (openSprints(), futureSprints()))"
                .format(self.project_id))
        else:
            sprint = self.current_sprint_id if not sprint else sprint
            search_query = 'sprint = {} '.format(sprint)
        if not assignee:
            assignee = "currentUser()"
        search_query += ' AND assignee = {}'.format(assignee)
        if status:
            search_query += ' AND status in ("{}")'.format(status)
        if text:
            search_query += ' AND (summary ~ "{}" OR description ~ "{}")'.format(text, text)
        return self.jira.search_issues(search_query)

    def get_my_issues(self):
        return self.search_issues()

    def get_worklog(self, issue):
        return self.jira.worklogs(issue.key)

    def get_todays_worklogs(self, issue_list):
        worklogs = []

        for issue in issue_list:
            for wl in self.get_worklog(issue):
                if iso_time_is_today(wl.started):
                    worklogs.append(wl)
        return worklogs

    def get_yesterdays_worklogs(self, issue_list):
        worklogs = []

        for issue in issue_list:
            for wl in self.get_worklog(issue):
                if iso_time_is_yesterday(wl.started):
                    worklogs.append(wl)
        return worklogs

    @staticmethod
    def edit_remaining_time(issue, time_string):
        """
        Set remaining time estimate on an issue.

        Keep originalEstimate and only edit remainingEstimate
        We need to pass both of them as not passing originalEstimate zeroes it.
        """
        try:
            original = issue.fields.timetracking.originalEstimate
        except AttributeError:
            print("Warning: issue had no timetracking field, using timeoriginalestimate field")
            original = friendly_worklog_time(issue.fields.timeoriginalestimate)
        f = IssueFields().timetracking(time_string, original)
        issue.update(**f.kwarg)

    @staticmethod
    def zero_remaining_time(issue):
        JiraWrapper.edit_remaining_time(issue, 0)

    def zero_remaining_work_done(self):
        """
        Find all "Done" issues assigned to me in the current sprint and 0 their time estimate.
        """
        issues = self.jira.search_issues(
            'sprint = {} AND assignee = currentUser() AND '
            'status = "Done" AND remainingEstimate > 0'.format(self.current_sprint_id)
        )

        for issue in issues:
            self.zero_remaining_time(issue)

    @staticmethod
    def normalize_name(txt):
        """
        Strip whitespace and switch to lowercase

        For example: "In Progress" becomes "inprogress"
        """
        return txt.replace(' ', '').lower()

    @property
    def component_labels_map(self):
        try:
            lowercase_data = {
                k.lower(): [l.lower() for l in v]
                for k, v in self._component_labels_map.items()
            }
        except KeyError:
            lowercase_data = {}
        return lowercase_data

    def find_component(self, txt):
        """
        Find component whose name or id matches 'txt', case insensitive

        Args:
          txt: str or int

        Returns:
          tuple of (component_name, component_id)
        """
        components = self.jira.project_components(self.project_id)
        for c in components:
            if (str(txt).isdigit and str(c.id) == str(txt)) or str(txt).lower() in c.name.lower():
                return c.name, c.id
        raise ValueError("Unable to find component with text: ", str(txt))

    def _check_comp_labels(self, component, labels):
        if not component or not labels:
            return
        if self.label_check:
            comp_lower = component.lower()
            labels_lower = [l.lower() for l in labels]
            if comp_lower in self.component_labels_map:
                for l in labels_lower:
                    if l not in self.component_labels_map[comp_lower]:
                        raise InvalidLabelError(component, l)

    def update_component(self, issue, component_name):
        server_side_name, _ = self.find_component(component_name)
        f = IssueFields().component(server_side_name)
        issue.update(**f.kwarg)

    @staticmethod
    def get_component(issue):
        if len(issue.fields.components) > 0:
            return issue.fields.components[0].name
        else:
            return None

    def update_labels(self, issue, labels):
        f = IssueFields().labels(labels)

        if hasattr(issue, 'components') and len(issue.components) > 0:
            self._check_comp_labels(issue.components[0].name, labels)

        issue.update(**f.kwarg)

    def find_status_name(self, txt):
        """
        Find the server-side status name based on 'txt' input.

        Will search using 'normalized' strings -- e.g. whitepsace removed and lowercase

        This way if txt is 'inprogress' this matches to "In Progress"
        """
        txt = self.normalize_name(txt)
        statuses = self._jira.statuses()
        for s in statuses:
            if txt == self.normalize_name(s.name):
                return s.name
        return None

    def get_avail_statuses(self, issue):
        """
        Find available status transitions for the given issue

        Builds a list of dicts, each dict contains:
           name: normalized name of the status, e.g. "inprogress"
           id: server-side if of the status
           friendly_name: the display name, e.g. "In Progress"
           local_num: the idx of this status, used for local selection in the CLI prompts
        """
        avail_statuses = [
            {
                'name': JiraWrapper.normalize_name(t['name']),  # used for name matching
                'id': t['id'],
                'friendly_name': t['name'],
            } for t in self.jira.transitions(issue) if 'Parallel Team' not in t['name']
        ]
        avail_statuses.sort(key=lambda s: s['name'])
        for idx, status in enumerate(avail_statuses):
            status['local_num'] = idx + 1
        return avail_statuses

    @staticmethod
    def get_avail_status_id(avail_statuses, txt):
        """
        Given a string name for an issue status, find the server-side id that correlates with it.

        Args:
            avail_statuses: list of avail status info return by JiraWrapper.get_avail_statuses()
            txt: string for status, e.x. "in progress" or "inprogress"
                 can also be a number which matches the status 'local_num' in avail_statuses

        Returns:
            status ID or None
        """
        for s in avail_statuses:
            normalized_name = JiraWrapper.normalize_name(txt)
            if normalized_name == s['name'] or (txt.isdigit() and int(txt) == s['local_num']):
                return s['id']
        return None

    def create_issue(self, summary, details=None, component=None,
                     labels=None, assignee=None, sprint=None, timeleft=None,
                     issuetype='Task', force_labels=False):
        """
        Create an issue (by default, a Story) in the agile sprint.

        Args:
          summary (str): issue title/summary
          details (str): detailed issue description
          component (str): component name
          labels (list of str): labels
          assignee (str): user id of assignee
          sprint (str): sprint name, sprint number, or 'backlog'. Default is current sprint
          timeleft (str): estimated time remaining (e.g. 2h30m)
          issueype (str): issue type, default is "Story", you likely won't change this.
          force_labels (boolean): don't check if labels/components are valid

        Returns:
          The newly created JIRA.Issue resource
        """

        if labels and not isinstance(labels, list):
            raise TypeError("labels must be a list")

        if not sprint:
            sprint_id = self.current_sprint_id
        else:
            _, sprint_id = self.find_sprint(sprint)

        if not force_labels:
            self._check_comp_labels(component, labels)

        f = IssueFields()
        comp_name_server_side, _ = self.find_component(component)
        f.summary(summary) \
            .description(details) \
            .component(comp_name_server_side) \
            .labels(labels) \
            .project(id=self.project_id) \
            .issuetype(issuetype) \
            .timetracking(timeleft, timeleft)
        
        new_issue = self.jira.create_issue(**f.kwarg)
        
        if assignee:
            self.jira.assign_issue(new_issue.key, assignee)

        if sprint == "backlog":
            self.jira.move_to_backlog([new_issue.key])
        else:
            self.jira.add_issues_to_sprint(sprint_id, [new_issue.key])

    def init(self):
        """Initialize all properties in one shot so it doesn't have to be done later."""
        # Note that these init self.jira too...
        print("Connecting to JIRA & gathering some info ...")
        try:
            self.userid
        except JIRAError as e:
            if 'CAPTCHA_CHALLENGE' in e.text:
                print("Your userID currently requires answering a CAPTCHA to login via basic auth")
                raw_input("Open {} in a browser, log in there to answer the CAPTCHA\n"
                           "Hit 'enter' here when done.".format(self.jira_url))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # Hide jira greenhopper API warnings
            print("UserID:", self.userid)
            print("Project ID:", self.project_id)
            print("Board ID:", self.board_id)
            print("Current sprint name:", self.current_sprint_name)
            print("Current sprint ID:", self.current_sprint_id)



"""
Contains utils such as update checker

"""
from __future__ import print_function

from distutils.version import StrictVersion
import json
import pkg_resources
import requests


PYPI_URL = 'http://pypi.python.org/pypi/jiraprompt/json'


def _compare_version(pypi_version):
    pypi_version = StrictVersion(pypi_version)
    try:
        local_version = pkg_resources.get_distribution('jiraprompt').version
    except pkg_resources.DistributionNotFound:
        local_version = '0.0.0'

    try:
        my_version = StrictVersion(local_version)
    except ValueError:
        print(
            'Version \'{}\' seems to be a dev version, assuming up-to-date'
            .format(local_version)
        )
        my_version = StrictVersion('999.999.999')

    if my_version < pypi_version:
        print(
            ' '
            'There is a new version available! (yours: {}, available: {})'
            ' '
            'Upgrade with:'
            '    pip install --upgrade jiraprompt'
            ' '
        )
    else:
        print('Up-to-date!')


def check_pypi():
    print('\nChecking pypi for latest release...')

    pkg_data = {}
    try:
        response = requests.get(PYPI_URL, timeout=5)
        response.raise_for_status()
        pkg_data = response.json()
    except requests.exceptions.Timeout:
        print('Unable to reach pypi quickly, giving up.')
    except requests.exceptions.HTTPError as e:
        print('Error response from pypi: ', e.errno, e.message)
    except ValueError:
        print('Response was not valid json, giving up.')
    
    try:
        pypi_version = pkg_data['info']['version']
    except KeyError:
        print('Unable to parse version info from pypi')
    else:
        _compare_version(pypi_version)
    print("\n")



import re
from datetime import datetime, timedelta

import editor
import iso8601
from dateutil import tz, parser


def editor_preserve_comments(default_text):
    """
    Open pyeditor and preserves comments. Does some encoding stuff.
    """
    if not isinstance(default_text, bytes):
        default_text = default_text.encode("utf-8")
    edited_text = editor.edit(contents=default_text)
    if not isinstance(edited_text, str):
        edited_text = edited_text.decode("utf-8")
    return edited_text


def editor_ignore_comments(default_text):
    """
    Open pyeditor but ignore lines starting with "#" when text is returned.

    :param default_text:
    :return:
    """
    if not isinstance(default_text, bytes):
        default_text = default_text.encode("utf-8")
    edited_text = editor.edit(contents=default_text)
    if not isinstance(edited_text, str):
        edited_text = edited_text.decode("utf-8")
    lines = edited_text.split('\n')
    return "\n".join(line for line in lines if not line.lstrip().startswith("#"))


def sanitize_worklog_time(s):
    """
    Convert a time string entered by user
    to jira-acceptable format for issue time tracking
    """
    s = s.replace(' ', '')

    def get_number_before(letter):
        number = 0
        try:
            regex_str = r'\D*(\d*)\s*{}.*'.format(letter)
            number = re.findall(regex_str, s)[0]
        except (AttributeError, IndexError):
            pass
        return number

    days = get_number_before('d')
    hours = get_number_before('h')
    mins = get_number_before('m')
    secs = get_number_before('s')

    new_s = ""
    new_s += days + "d " if days else ""
    new_s += hours + "h " if hours else ""
    new_s += mins + "m " if mins else ""
    new_s += secs + "s " if secs else ""
    if new_s:
        return new_s
    else:
        # user might not have specified any strings at all,
        # just pass along the int
        return s


def friendly_worklog_time(seconds):
    """
    https://stackoverflow.com/questions/775049/how-to-convert-seconds-to-hours-minutes-and-seconds

    :param seconds:
    :return:
    """
    if not seconds:
        string = "0m"
    else:
        m, s = divmod(int(seconds), 60)
        h, m = divmod(m, 60)
        string = ""
        string += "{}h".format(h) if h else ""
        string += "{}m".format(m) if m else ""
        string += "{}s".format(s) if s else ""
    return string


def iso_to_datetime(string):
    tz_utc = tz.tzutc()
    tz_local = tz.tzlocal()
    utc_datetime = iso8601.parse_date(string)
    utc_datetime = utc_datetime.replace(tzinfo=tz_utc)
    return utc_datetime.astimezone(tz_local)


# For now, instead of using hard-coded datetime formats
# we'll use %c for datetime->string and for string->datetime
# just use dateutil's parser

# TIME_FORMAT = '%a %d %b %Y %I:%M:%S %p %Z'


def iso_to_ctime_str(string):
    datetime_object = iso_to_datetime(string)
    return datetime_object.strftime('%c')


def ctime_str_to_datetime(datetime_string):
    return parser.parse(datetime_string)


def ctime_str_to_iso(datetime_string):
    return ctime_str_to_datetime(datetime_string).isoformat()


def iso_time_is_today(string):
    datetime_object = iso_to_datetime(string)
    return datetime.today().date() == datetime_object.date()


def iso_time_is_yesterday(string):
    datetime_object = iso_to_datetime(string)
    return (datetime.today().date() - timedelta(1)) == datetime_object.date()


from __future__ import print_function

import argparse
import os
import sys

from pathlib2 import Path
import prompter
import yaml

from .common import editor_preserve_comments
from .prompt import MainPrompt
from .res import get_default_config, get_default_labels, get_ascii_art
from .utils.update_check import check_pypi


DEFAULT_CONFIG_FILE = 'config.yaml'
DEFAULT_LABELS_FILE = 'labels.yaml'


def _get_config_path():
    xdg_config_home = os.environ.get('XDG_CONFIG_HOME')
    if xdg_config_home:
        config_home = Path(xdg_config_home)
    else:
        config_home = Path.home().joinpath('.config')
        
    return config_home.joinpath('jiraprompt')


def _write_config_file(filename, txt):
    path = _get_config_path().joinpath(filename)
    path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
    path.write_text(txt.decode('utf8'))
    path.chmod(0o600)
    return path


def _create_config_files():
    new_config = editor_preserve_comments(get_default_config())
    config_path = _write_config_file(DEFAULT_CONFIG_FILE, new_config)

    new_labels_config = editor_preserve_comments(get_default_labels())
    labels_path = _write_config_file(DEFAULT_LABELS_FILE, new_labels_config)

    print('Writing config to {}'.format(config_path))
    print('Writing labels config to {}'.format(labels_path))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config-file', type=str)
    parser.add_argument('--labels-file', type=str)
    args, unknown_args = parser.parse_known_args()

    if args.config_file:
        config_path = Path(args.config_file)
    else:
        config_path = _get_config_path().joinpath(DEFAULT_CONFIG_FILE)
        if not config_path.exists():
            print('It looks like you have no config created.\n')
            if prompter.yesno('Create one now?'):
                _create_config_files()

    if not config_path.exists():
        print('ERROR: No valid config file found.')
        sys.exit(1)

    if args.labels_file:
        labels_path = Path(args.labels_file)
    else:
        labels_path = _get_config_path().joinpath(DEFAULT_LABELS_FILE)
    if not labels_path.exists():
        print(
            'WARNING: Labels config file not found at \'{}\''
            .format(labels_path)
        )
        labels_path = ''

    # print welcome msg
    print(get_ascii_art().decode('utf8'))

    # check for updates if enabled
    with config_path.open() as f:
        cfg = yaml.safe_load(f)
    if cfg.get('check_for_updates', True):
        check_pypi()

    sys.argv = sys.argv[:1] + unknown_args

    main_prompt = MainPrompt(
        config_file=str(config_path), labels_file=str(labels_path))
    main_prompt.cmdloop()


if __name__ == '__main__':
    main()
